Scriptname BCP_MainScript extends ReferenceAlias  


;-------------------------Properties & variables-------------------------

Quest Property BCP_ItemAddedQuest Auto
Quest Property BCP_SavedValuesQuest Auto
Quest Property BCP_MCMQuest Auto

Bool Property debugVerbosity = False Auto

int Property SKSEVersion Auto

GlobalVariable Property BCP_InstalledVersionGV Auto
GlobalVariable Property BCP_VersionGV Auto

Message Property BCP_InitializeMsg Auto
Message Property BCP_SKSEMissingMsg Auto
Message Property BCP_SKSEMismatchMsg Auto
Message Property BCP_ErrorCalculateMsg Auto
Message Property BCP_CalculateBeginMsg Auto
Message Property BCP_CalculateEndMsg Auto
Message Property BCP_ErrorResetMsg Auto

GlobalVariable Property BCP_MaxSavedPotionsGV Auto

GlobalVariable Property BCP_RecalculateGV Auto
GlobalVariable Property BCP_ResetGV Auto

Bool maintenanceBusy = False


GlobalVariable Property BCP_MatchModelGV Auto
GlobalVariable Property BCP_MatchNameGV Auto
GlobalVariable Property BCP_EffectsInNameGV Auto
GlobalVariable Property BCP_EffectsAbbreviationGV Auto
GlobalVariable Property BCP_CalculateImpureGoldGV Auto
GlobalVariable Property BCP_BrewedSuffixGV Auto
GlobalVariable Property BCP_MixedSuffixGV Auto
GlobalVariable Property BCP_ImpureSuffixGV Auto
GlobalVariable Property BCP_NameTrimmingGV Auto
GlobalVariable Property BCP_MaxNameLengthGV Auto
GlobalVariable Property BCP_CalculateXPGV Auto

Perk Property BCP_XPGainStopPerk Auto
GlobalVariable Property XPMultGV Auto
Float totalXPGained = 0.0

Bool Property matchTextureSets = False Auto
Bool Property multiEffectMatch = False Auto

Potion Property defaultPoison Auto
Potion Property defaultPotion Auto

Keyword Property MagicAlchBeneficial Auto
Keyword Property MagicAlchHarmful Auto

string[] Property effectNamesArray Auto
string[] Property abbrvEffectNamesArray Auto

String Property brewedSuffixString Auto
String Property mixedSuffixString Auto
String Property impureSuffixString Auto

FormList Property BCP_CraftedPotionsFormList Auto
FormList Property BCP_AlchEffects Auto
FormList Property BCP_GroupedAlchMultiEffects Auto
FormList Property BCP_GroupedPotions Auto
FormList Property BCP_GroupedMultiPotions Auto






;-------------------------Events-------------------------

Event OnInit()
	Debug.TraceConditional(self + " OnInit event", debugVerbosity)
	If SKSE.GetVersionRelease() > 0
		RegisterForMenu("Crafting Menu")
		BCP_InstalledVersionGV.SetValue(BCP_VersionGV.GetValue())
		BCP_InitializeMsg.Show()
		If !maintenanceBusy
			Maintenance()
		EndIf
	Else
		BCP_SKSEMissingMsg.Show()
	EndIf
EndEvent
 
 
Event OnPlayerLoadGame()
 	Debug.TraceConditional( self + " OnPlayerLoadGame event start time:" + Utility.GetCurrentRealTime(), debugVerbosity)
	If !BCP_VersionGV || BCP_VersionGV.GetValue() != BCP_InstalledVersionGV.GetValue()
		Debug.TraceConditional(self + " Update detected", debugVerbosity)
		BCP_MCMQuest.Stop()
		BCP_MCMQuest.Start()
		(BCP_MCMQuest as BCP_MCMScript).ForcePageReset()
		self.GetOwningQuest().Stop()
		self.GetOwningQuest().Start()
	Elseif !maintenanceBusy
		Maintenance()
	EndIf
	Debug.TraceConditional( self + " OnPlayerLoadGame event end time:" + Utility.GetCurrentRealTime(), debugVerbosity)
 EndEvent
 
 
Event OnMenuOpen(String MenuName)
	If MenuName == "Crafting Menu" && !maintenanceBusy
		Debug.TraceConditional(self + " Crafting menu opened", debugVerbosity)
		BCP_ItemAddedQuest.Start()
		If BCP_CalculateImpureGoldGV.GetValue() == 1.0 && BCP_CalculateXPGV.GetValue() == 1.0
			Game.GetPlayer().AddPerk(BCP_XPGainStopPerk)
		EndIf
	EndIf
EndEvent


Event OnMenuClose(String MenuName)
	If MenuName == "Crafting Menu" && !maintenanceBusy
		Debug.TraceConditional(self + " Crafting menu closed", debugVerbosity)
		BCP_ItemAddedQuest.Stop()
		If Game.GetPlayer().HasPerk(BCP_XPGainStopPerk)
			Game.GetPlayer().RemovePerk(BCP_XPGainStopPerk)
		EndIf
		If BCP_CalculateImpureGoldGV.GetValue() == 1.0 && BCP_CalculateXPGV.GetValue() == 1.0
			If XPMultGV
				totalXPGained = totalXPGained * XPMultGV.GetValue()
			EndIf
			If totalXPGained > 0.0
				Game.AdvanceSkill("Alchemy", totalXPGained)
				Debug.TraceConditional( self + " Xp added to player: " + totalXPGained, debugVerbosity)
			EndIf
		totalXPGained = 0.0
		EndIf
	EndIf
EndEvent






;-------------------------Maintenance Functions-------------------------

Function Maintenance()
	maintenanceBusy = True

	If BCP_CalculateImpureGoldGV.GetValue() == 1.0 && SKSEVersion != SKSE.GetVersionRelease()
		BCP_SKSEMismatchMsg.Show()
		Debug.TraceConditional(self + "SKSE version mismatch detected. Installed version: " + SKSE.GetVersionRelease(), debugVerbosity)
	EndIf	

	Debug.TraceConditional(self + " Potions formlist:" + BCP_CraftedPotionsFormList.ToArray(), debugVerbosity)
	Debug.TraceConditional(self + " Base potions array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, debugVerbosity)
	Debug.TraceConditional(self + " Names array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, debugVerbosity)
	Debug.TraceConditional(self + " Prices array" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, debugVerbosity)
	
	CleanSavedData()
	
	Debug.TraceConditional(self + " Cleaned potions formlist:" + BCP_CraftedPotionsFormList.ToArray(), debugVerbosity)
	Debug.TraceConditional(self + " Cleaned base potions array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, debugVerbosity)
	Debug.TraceConditional(self + " Cleaned names array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, debugVerbosity)
	Debug.TraceConditional(self + " Cleaned prices array" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, debugVerbosity)

	If BCP_ResetGV.GetValue() == 1.0
		Debug.TraceConditional(self + " Saved data erased by user request.", debugVerbosity)
		EraseSavedData()
		BCP_ResetGV.SetValue(0.0)
	ElseIf BCP_RecalculateGV.GetValue() == 1.0
		Debug.TraceConditional(self + " Potions values recalculated by user request.", debugVerbosity)
		RecalculatePotions()
		BCP_RecalculateGV.SetValue(0.0)
		Debug.TraceConditional(self + " Recalculated potions formlist:" + BCP_CraftedPotionsFormList.ToArray(), debugVerbosity)
		Debug.TraceConditional(self + " Recalculated base potions array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, debugVerbosity)
		Debug.TraceConditional(self + " Recalculated names array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, debugVerbosity)
		Debug.TraceConditional(self + " Recalculated prices array" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, debugVerbosity)
		If !IsSavedDataValid()
			Debug.TraceConditional(self + " Saved arrays do not match formlist length after recalculation, should reset.", debugVerbosity)
			BCP_ErrorResetMsg.Show()
			EraseSavedData()
		EndIf
	ElseIf !IsSavedDataValid()
		Debug.TraceConditional(self + " Saved arrays do not match formlist length, should recalculate.", debugVerbosity)
		BCP_ErrorCalculateMsg.Show()
		RecalculatePotions()
		Debug.TraceConditional(self + " Recalculated potions formlist:" + BCP_CraftedPotionsFormList.ToArray(), debugVerbosity)
		Debug.TraceConditional(self + " Recalculated base potions array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, debugVerbosity)
		Debug.TraceConditional(self + " Recalculated names array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, debugVerbosity)
		Debug.TraceConditional(self + " Recalculated prices array" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, debugVerbosity)
		If !IsSavedDataValid()
			Debug.TraceConditional(self + " Saved arrays do not match formlist length after recalculation, should reset.", debugVerbosity)
			BCP_ErrorResetMsg.Show()
			EraseSavedData()
		EndIf
	Else
		LoadSavedPotions()
	EndIf
	
	maintenanceBusy = False
EndFunction



Function CleanSavedData()
;removes oldest data if more than set maximum is saved.
;removes none values from potion formlist and removes corresponding values in saved data.
	int formListLength = BCP_CraftedPotionsFormList.GetSize()
	bool shouldClean = False
	int dataIndex = 0
	int MaxSavedPotions = BCP_MaxSavedPotionsGV.GetValue() as int
	If formListLength > MaxSavedPotions
		dataIndex = formListLength - MaxSavedPotions
		shouldClean = True
	EndIf
	int index = dataIndex
	While !shouldClean && index < formListLength
		If !(BCP_CraftedPotionsFormList.GetAt(index) as Potion)
			shouldClean = True ;at least one none value is present in crafted potions formlist.
		EndIf
		index = index + 1
	EndWhile
	If shouldClean
		Form[] potionsArray = BCP_CraftedPotionsFormList.ToArray()
		BCP_CraftedPotionsFormList.Revert()
		Form[] basePotionsArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions
		Form[] cleanedBasePotionsArray
		string[] namesArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames
		string[] cleanedNamesArray
		int[] pricesArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices
		int[] cleanedpricesArray
		Potion currentPotion
		While dataIndex < formListLength
			If potionsArray[dataIndex] as Potion
				currentPotion = potionsArray[dataIndex] as Potion
				BCP_CraftedPotionsFormList.AddForm(currentPotion)
				cleanedBasePotionsArray = AddToFormArray(cleanedBasePotionsArray, basePotionsArray[dataIndex])
				cleanedNamesArray = AddToStringArray(cleanedNamesArray, namesArray[dataIndex])
				cleanedPricesArray = AddToIntArray(cleanedPricesArray, pricesArray[dataIndex])
			Else
				Debug.TraceConditional(self + " found none form in formlist at index:" + dataIndex, debugVerbosity)
			EndIf
			dataIndex = dataIndex + 1
		EndWhile
		(BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions = cleanedBasePotionsArray
		(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames = cleanedNamesArray
		(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices = cleanedPricesArray
	EndIf
EndFunction



Function LoadSavedPotions()
	bool nameTrimming = BCP_NameTrimmingGV.GetValue() == 1.0
	int maxNameLength = BCP_MaxNameLengthGV.GetValue() as int
	
	Form[] basePotionsArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions
	string[] savedNamesArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames
	int[] savedPricesArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices
	
	int intNumberToLoad = BCP_CraftedPotionsFormList.GetSize()
	int index = 0
	Potion currentPotion
	string potionName
	While index < intNumberToLoad
		currentPotion = BCP_CraftedPotionsFormList.GetAt(index) as Potion
		If currentPotion
			If nameTrimming && StringUtil.GetLength(savedNamesArray[index]) > maxNameLength + 2
				potionName = (StringUtil.Substring(savedNamesArray[index], 0, maxNameLength) + "..")
			Else
				potionName = savedNamesArray[index]
			EndIf
			ApplyValues(currentPotion, basePotionsArray[index] as Potion, potionName, savedPricesArray[index])
		EndIf
		index = index + 1
	EndWhile
EndFunction



Bool Function IsSavedDataValid()
	;checks if every formlist/array saved have the same length.
	;If not, something went wrong and script should recalculate potions or even reset if problem persists.
	Bool dataIsValid = False
	Form[] basePotionsArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions
	string[] savedNamesArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames
	int[] savedPricesArray = (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices
	int formListLength = BCP_CraftedPotionsFormList.GetSize()
	If formListLength == basePotionsArray.Length
		If basePotionsArray.length == savedNamesArray.length
			If savedNamesArray.length == savedPricesArray.length
				dataIsValid = True
			EndIf
		EndIf
	EndIf
	return dataIsValid
EndFunction



Function RecalculatePotions()
	BCP_CalculateBeginMsg.Show()
	int formListLength = BCP_CraftedPotionsFormList.GetSize()
	;(BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions = Utility.CreateFormArray(formListLength, None)
	;(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames = Utility.CreateStringArray(formListLength, "")
	;(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices = Utility.CreateIntArray(formListLength, 0)
	Form[] emptyFormArray
	String[] emptyStringArray
	Int[] emptyIntArray
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions = emptyFormArray
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames = emptyStringArray
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices = emptyIntArray
	int index = 0
	Potion craftedPotion
	While index < formListLength
		craftedPotion = BCP_CraftedPotionsFormList.GetAt(index) as Potion
		If craftedPotion
			NewPotionCrafted(craftedPotion)
		EndIf
		index = index + 1
	EndWhile
	BCP_CalculateEndMsg.Show()
EndFunction



Function EraseSavedData() ;something is wrong, erasing all data
	BCP_CraftedPotionsFormList.Revert()
	Form[] emptyFormArray
	String[] emptyStringArray
	Int[] emptyIntArray
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions = emptyFormArray
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames = emptyStringArray
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices = emptyIntArray
EndFunction



Function SaveData(Potion craftedPotion, form basePotion, string potionName, int PotionValue)
	int formListLength = BCP_CraftedPotionsFormList.GetSize()
	int index = 0
	If !BCP_CraftedPotionsFormList.HasForm(craftedPotion)
		BCP_CraftedPotionsFormList.AddForm(craftedPotion)
	EndIf
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions = AddToFormArray((BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, basePotion)
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames = AddToStringArray((BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, potionName)
	(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices = AddToIntArray((BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, PotionValue)
EndFunction






;-------------------------Arrays Manipulation Functions-------------------------

string[] Function AddToStringArray(string[] stringArray, string newElement)
	int arrayLength = stringArray.length
	string[] newArray = Utility.CreateStringArray(arrayLength + 1, "")
	int index = 0
	While index < arrayLength
		newArray[index] = stringArray[index]
		index = index + 1
	EndWhile
	newArray[index] = newElement
	return newArray
EndFunction


int[] Function AddToIntArray(int[] intArray, int newElement)
	int arrayLength = intArray.length
	int[] newArray = Utility.CreateIntArray(arrayLength + 1, 0)
	int index = 0
	While index < arrayLength
		newArray[index] = intArray[index]
		index = index + 1
	EndWhile
	newArray[index] = newElement
	return newArray
EndFunction


Form[] Function AddToFormArray(Form[] formsArray, Form newElement)
	int arrayLength = formsArray.length
	Form[] newArray = Utility.CreateFormArray(arrayLength + 1, None)
	int index = 0
	While index < arrayLength
		newArray[index] = formsArray[index]
		index = index + 1
	EndWhile
	newArray[index] = newElement
	return newArray
EndFunction



string[] Function PushToBottomStringArray(string[] stringArray, int elementIndex)
	int arrayLength = stringArray.length
	string[] newArray = Utility.CreateStringArray(arrayLength, "")
	int index = 0
	While index < elementIndex
		newArray[index] = stringArray[index]
		index = index + 1
	EndWhile
	While index < arrayLength - 1
		newArray[index] = stringArray[index + 1]
		index = index + 1
	EndWhile
	newArray[index] = stringArray[elementIndex]
	return newArray
EndFunction


int[] Function PushToBottomIntArray(int[] intArray, int elementIndex)
	int arrayLength = intArray.length
	int[] newArray = Utility.CreateIntArray(arrayLength, 0)
	int index = 0
	While index < elementIndex
		newArray[index] = intArray[index]
		index = index + 1
	EndWhile
	While index < arrayLength - 1
		newArray[index] = intArray[index + 1]
		index = index + 1
	EndWhile
	newArray[index] = intArray[elementIndex]
	return newArray
EndFunction


form[] Function PushToBottomFormArray(form[] formsArray, int elementIndex)
	int arrayLength = formsArray.length
	form[] newArray = Utility.CreateFormArray(arrayLength, None)
	int index = 0
	While index < elementIndex
		newArray[index] = formsArray[index]
		index = index + 1
	EndWhile
	While index < arrayLength - 1
		newArray[index] = formsArray[index + 1]
		index = index + 1
	EndWhile
	newArray[index] = formsArray[elementIndex]
	return newArray
EndFunction






;-------------------------Potions Handling Functions-------------------------

Function KnownPotionCrafted(Potion craftedPotion)
;if a already known potion is crafted, it should be pushed to the bottom of the formlist (the higher in the formlist, the sooner it gets deleted).
	Debug.TraceConditional( self + " KnownPotionCrafted function start time:" + Utility.GetCurrentRealTime(), debugVerbosity)
	
	int index = 0
	int formListlength = 0
	index = BCP_CraftedPotionsFormList.Find(craftedPotion)
	formListlength = BCP_CraftedPotionsFormList.GetSize()
	If index == formListLength - 1
		Debug.TraceConditional(self + " crafted potion is already at the bottom of the list.", debugVerbosity)
		
	ElseIf index > -1
		Debug.TraceConditional(self + " Potions formlist:" + BCP_CraftedPotionsFormList.ToArray(), debugVerbosity)
		Debug.TraceConditional(self + " Base potions array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, debugVerbosity)
		Debug.TraceConditional(self + " Names array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, debugVerbosity)
		Debug.TraceConditional(self + " Prices array" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, debugVerbosity)
		
		form[] potionsArray = BCP_CraftedPotionsFormList.ToArray()
		potionsArray = PushToBottomFormArray(potionsArray, index)
		BCP_CraftedPotionsFormList.Revert()
		BCP_CraftedPotionsFormList.AddForms(potionsArray)
		(BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions = PushToBottomFormArray((BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, index)
		(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames = PushToBottomStringArray((BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, index)
		(BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices = PushToBottomIntArray((BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, index)
		
		Debug.TraceConditional(self + " Updated potions formlist:" + BCP_CraftedPotionsFormList.ToArray(), debugVerbosity)
		Debug.TraceConditional(self + " Updated base potions array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).basePotions, debugVerbosity)
		Debug.TraceConditional(self + " Updated names array:" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedNames, debugVerbosity)
		Debug.TraceConditional(self + " Updated prices array" + (BCP_SavedValuesQuest as BCP_SavedValuesScript).savedPrices, debugVerbosity)
	EndIf
	
	Debug.TraceConditional( self + " KnownPotionCrafted function end time:" + Utility.GetCurrentRealTime(), debugVerbosity)
EndFunction



Function NewPotionCrafted(Potion craftedPotion)
	Debug.TraceConditional( self + " NewPotionCrafted function start time:" + Utility.GetCurrentRealTime(), debugVerbosity)

	;MCM options
	bool matchModel = BCP_MatchModelGV.GetValue() == 1.0
	bool matchName = BCP_MatchNameGV.GetValue() == 1.0
	bool effectsInName = BCP_EffectsInNameGV.GetValue() == 1.0
	bool effectsAbbreviation = BCP_EffectsAbbreviationGV.GetValue() == 1.0
	bool calculateImpureGold = BCP_CalculateImpureGoldGV.GetValue() == 1.0
	bool brewedSuffix = BCP_BrewedSuffixGV.GetValue() == 1.0
	bool mixedSuffix = BCP_MixedSuffixGV.GetValue() == 1.0
	bool impureSuffix = BCP_ImpureSuffixGV.GetValue() == 1.0
	bool nameTrimming = BCP_NameTrimmingGV.GetValue() == 1.0
	int maxNameLength = BCP_MaxNameLengthGV.GetValue() as int
	bool calculateXP = BCP_CalculateXPGV.GetValue() == 1.0
	Debug.TraceConditional(self + " matchModel:" + matchModel + " matchName:" + matchName + " effectsInName:" + effectsInName + " calculateImpureGold:" + calculateImpureGold, debugVerbosity)
	Debug.TraceConditional(self + " Suffix:" + brewedSuffix + mixedSuffix + impureSuffix, debugVerbosity)
	
	;potion info
	int craftedPotionNumberEffects = craftedPotion.GetNumEffects()
	bool craftedPotionIsPoison = craftedPotion.IsPoison()
	bool craftedPotionIsImpure = False
	If craftedPotionNumberEffects > 1
		craftedPotionIsImpure = craftedPotion.HasKeyword(MagicAlchBeneficial) && craftedPotion.HasKeyword(MagicAlchHarmful)
	EndIf
	
	;default/fallback values 
	Potion closestPotion
	If craftedPotionIsPoison
		closestPotion = defaultPoison
	Else
		closestPotion = defaultPotion
	EndIf
	string craftedPotionName = craftedPotion.GetName()
	string craftedPotionNameAbrv = craftedPotionName
	int craftedPotionValue = craftedPotion.GetGoldValue()
	
	;base potion and name
	If matchModel || matchName
		int craftedPotionPrimaryEffectIndex
		If !craftedPotionIsImpure
			craftedPotionPrimaryEffectIndex = craftedPotion.GetCostliestEffectIndex()
		Else ;costliest effect is problematic for impure potions, sometimes will a give positive effect for a poison (or negative for a potion)
			craftedPotionPrimaryEffectIndex = 0 ;seems like first effect dictates potion or poison
		EndIf
		closestPotion = FindClosestPotion(craftedPotion, craftedPotionPrimaryEffectIndex)
			If closestPotion
				If matchName
					craftedPotionName = closestPotion.GetName()
					Debug.TraceConditional(self + " closest name found:" + craftedPotionName, debugVerbosity)
				EndIf
			EndIf
	EndIf
	
	;Suffix
	If closestPotion
		string suffixString = ""
		If impureSuffix && craftedPotionIsImpure
			If brewedSuffix
				suffixString = brewedSuffixString + ", " + impureSuffixString
			Else
				suffixString = impureSuffixString
			EndIf
		ElseIf mixedSuffix && !craftedPotionIsImpure && craftedPotionNumberEffects > 1
			If !(matchName && craftedPotionNumberEffects <= closestPotion.GetNumEffects()) ;if potion matches exactly a multi-effect loot potion there is no need for the "mixed" suffix
				If brewedSuffix
					suffixString = brewedSuffixString + ", " + mixedSuffixString
				Else
					suffixString = mixedSuffixString
				EndIf
			EndIf
		ElseIf brewedSuffix
			suffixString = brewedSuffixString
		EndIf
		If SuffixString != ""
			craftedPotionName = craftedPotionName + " (" + SuffixString + ")"
		EndIf
	EndIf

	;effects name & impure gold
	If (craftedPotionNumberEffects > 1 && effectsInName) || (craftedPotionIsImpure && calculateImpureGold)
		float craftedPotionBenValue = 0.0
		float craftedPotionHarmValue = 0.0
		int index = craftedPotionNumberEffects
		string effectsString = ")"
		string effectName
		int effectNameIndex
		MagicEffect currentMagicEffect
		float currentEffectBaseCost
		float currentEffectMagnitude
		float currentEffectDuration
		float currentEffectCost
		While index > 0
			currentMagicEffect = craftedPotion.GetNthEffectMagicEffect(index - 1)
			If effectsInName
				effectName = currentMagicEffect.GetName()
				If effectsAbbreviation
					effectNameIndex = effectNamesArray.find(effectName)
					If effectNameIndex >= 0
						effectName = abbrvEffectNamesArray[effectNameIndex]
					EndIf
				EndIf
				effectsString = effectName + effectsString
				If index == 1
					effectsString = " (" + effectsString
				ElseIf index == craftedPotionNumberEffects
					effectsString = " & " + effectsString
				ElseIf index > 1
					effectsString = ", " + effectsString
				EndIf
			EndIf
			If craftedPotionIsImpure && calculateImpureGold
				currentEffectBaseCost = currentMagicEffect.GetBaseCost()
				currentEffectMagnitude = craftedPotion.GetNthEffectMagnitude(index - 1)
				currentEffectDuration = craftedPotion.GetNthEffectDuration(index - 1)
				currentEffectCost = CalculateEffectCost(currentEffectBaseCost, currentEffectMagnitude, currentEffectDuration)
				If currentMagicEffect.HasKeyword(MagicAlchBeneficial)
					craftedPotionBenValue = craftedPotionBenValue + currentEffectCost
				ElseIf currentMagicEffect.HasKeyword(MagicAlchHarmful)
					craftedPotionHarmValue = craftedPotionHarmValue + currentEffectCost
				EndIf
			EndIf
			index = index - 1
		EndWhile
		If effectsInName
			craftedPotionName = craftedPotionName + effectsString
			Debug.TraceConditional(self + " new name:" + craftedPotionName, debugVerbosity)
		EndIf
		If craftedPotionIsImpure && calculateImpureGold
			Debug.TraceConditional(self + " ben effects value:" + craftedPotionBenValue + " neg effects value:" + craftedPotionHarmValue, debugVerbosity)
			float calculatedGold
			If craftedPotionIsPoison
				calculatedGold = craftedPotionHarmValue - craftedPotionBenValue
			Else
				calculatedGold = craftedPotionBenValue - craftedPotionHarmValue
			EndIf
			If calculatedGold < 0.0
				calculatedGold = 0.0
			EndIf
			craftedPotionValue = Math.Floor(calculatedGold)
		EndIf
	EndIf

	;XP Calculation
	If calculateImpureGold && calculateXP
		totalXPGained = totalXPGained + craftedPotionValue
	EndIf
	
	;name trimming
	If nameTrimming && StringUtil.GetLength(craftedPotionName) > maxNameLength + 2
		craftedPotionNameAbrv = StringUtil.Substring(craftedPotionName, 0, maxNameLength) + ".."
	Else
		craftedPotionNameAbrv = craftedPotionName
	EndIf
	
	;Applying values to potion
	ApplyValues(craftedPotion, closestPotion, craftedPotionNameAbrv, craftedPotionValue)
	
	;Saving values for future game load
	SaveData(craftedPotion, closestPotion, craftedPotionName, craftedPotionValue)
	
	Debug.TraceConditional( self + " NewPotionCrafted function end time:" + Utility.GetCurrentRealTime(), debugVerbosity)
EndFunction



Function ApplyValues(Potion craftedPotion, Potion basePotion, string potionName, int potionValue)
	If BCP_MatchModelGV.GetValue() == 1.0 && basePotion
		craftedPotion.SetWorldModelPath(basePotion.GetWorldModelPath())
		If matchTextureSets
			int numTextureSets = 0
			numTextureSets = basePotion.GetWorldModelNumTextureSets()
			int textureSetIndex = 0
			While textureSetIndex < numTextureSets
				craftedPotion.SetWorldModelNthTextureSet(basePotion.GetWorldModelNthTextureSet(textureSetIndex), textureSetIndex)
				textureSetIndex = textureSetIndex + 1
			EndWhile
		EndIf
	EndIf
	craftedPotion.SetName(potionName)
	If BCP_CalculateImpureGoldGV.GetValue() == 1.0
		DP_Extender.SetAlchAutoCalcFlag(craftedPotion, false)
		craftedPotion.SetGoldValue(potionValue)
	EndIf
EndFunction



float Function CalculateEffectCost(float baseCost, float magnitude, float duration)
	float cost
	If magnitude < 1.0
		magnitude = 1.0
	EndIf
	If duration == 0.0
		cost = baseCost * Math.Pow(magnitude, 1.1)
	Else
		cost = baseCost * Math.Pow(magnitude, 1.1) * Math.Pow(duration/10, 1.1)
	EndIf
	return cost
EndFunction




FormList Function FindMatchingSingleEffectPotionsFormList(MagicEffect primaryEffect, FormList effectsFormList, FormList groupedPotionsFormList)
	FormList matchingPotionsFormList = None
	int effectIndex = effectsFormList.Find(primaryEffect) ;searching in formlist of possible magic effects
	If effectIndex > -1 ; -1 means no matching effect
		Debug.TraceConditional(self as string + primaryEffect + " primary effect was found in effects formlist at index: " + effectIndex, debugVerbosity)
		matchingPotionsFormList = groupedPotionsFormList.GetAt(effectIndex) as formlist ;formlist of formlist of potions, grouped by effect
	Else
		Debug.TraceConditional(self as string + primaryEffect + " primary effect was not found in effects formlist", debugVerbosity)
	EndIf
	Return matchingPotionsFormList
EndFunction



FormList Function FindMatchingMultiEffectPotionsFormList(Potion akPotion, FormList groupedEffectsFormList, FormList groupedPotionsFormList)
	FormList matchingPotionsFormList = None
	
	int groupedEffectsIndex = 0
	int groupedEffectsLength = groupedEffectsFormList.GetSize()
	int potionNumEffects = akPotion.GetNumEffects()
	While groupedEffectsIndex < groupedEffectsLength && !matchingPotionsFormList ;scanning a formlist of formlist of magic effects
		int potionEffectsIndex = 0
		FormList effectsFormList = groupedEffectsFormList.GetAt(groupedEffectsIndex) as FormList
		int formListNumEffects = effectsFormList.GetSize()
		int numMatchingEffect = 0
		While potionEffectsIndex < potionNumEffects && numMatchingEffect < formListNumEffects ;scanning a formlist of magic effects
			If effectsFormList.Find(akPotion.GetNthEffectMagicEffect(potionEffectsIndex)) != -1
				numMatchingEffect = numMatchingEffect + 1 ;counting the number of matching effects between the formlist and the potion
			EndIf
			potionEffectsIndex = potionEffectsIndex + 1
		EndWhile
		If numMatchingEffect == formListNumEffects ;all effects in the formlist are present in the potion, it's a multi-effects match
			Debug.TraceConditional(self as string + " multi effects match found at index: " + groupedEffectsIndex, debugVerbosity)
			matchingPotionsFormList = groupedPotionsFormList.GetAt(groupedEffectsIndex) as FormList
		EndIf
		groupedEffectsIndex = groupedEffectsIndex + 1
	EndWhile
	
	Return matchingPotionsFormList
EndFunction



int Function FindEffectIndexInPotion(magicEffect akEffect, Potion akPotion)
	int foundIndex = -1
	int index = 0
	int numEffects = akPotion.GetNumEffects()
	While index < numEffects && foundIndex == -1
		If akPotion.GetNthEffectMagicEffect(index) == akEffect
			foundIndex = index
		EndIf
		index = index + 1
	EndWhile
	return foundIndex
EndFunction



Potion Function FindClosestPotion(potion akPotion, int akPotionEffectIndex)
;finds closest matching potion (in terms of magnitude or duration depending on which one scales with power) in a two formlists of formlists of potions grouped by effects.
	potion closestPotion = None
	
	FormList potionsList = None
	If multiEffectMatch
		potionsList = FindMatchingMultiEffectPotionsFormList(akPotion, BCP_GroupedAlchMultiEffects, BCP_GroupedMultiPotions)
	EndIf
	If !potionsList
		potionsList = FindMatchingSingleEffectPotionsFormList(akPotion.GetNthEffectMagicEffect(akPotionEffectIndex), BCP_AlchEffects, BCP_GroupedPotions)
	EndIf
		
	int listLength = 0
	If potionsList
		listLength = potionsList.GetSize()
	EndIf
	If akPotion && potionsList && listLength > 0 ;0 means no leveled list potions with that effect
		int index = 0
		float potionPotency = 0.0
		closestPotion = potionsList.GetAt(index) as potion
		
		int closestPotionEffectIndex = FindEffectIndexInPotion(akPotion.GetNthEffectMagicEffect(akPotionEffectIndex), closestPotion)	
		If closestPotionEffectIndex == -1 ;some mods (like CACO) have effects in multi effects potions that are not in ingredients, thus not in crafted potions.
			Debug.TraceConditional(self + "primary effect not in multi-effect potion, looking for first matching effect.", debugVerbosity)
			int akPotionNumEffects = akPotion.GetNumEffects()
			akPotionEffectIndex = 0
			While akPotionEffectIndex < akPotionNumEffects && closestPotionEffectIndex == -1
				closestPotionEffectIndex =  FindEffectIndexInPotion(akPotion.GetNthEffectMagicEffect(akPotionEffectIndex), closestPotion)
				akPotionEffectIndex = akPotionEffectIndex + 1
			EndWhile
			Debug.TraceConditional(self + "matching effect found at index " + akPotionEffectIndex + " in crafted potion and at index " + closestPotionEffectIndex + " in loot potion " + closestPotion, debugVerbosity)
		EndIf
		
		If closestPotionEffectIndex != -1 && akPotion.GetNthEffectMagicEffect(akPotionEffectIndex).IsEffectFlagSet(0x00200000) ;effect scales with magnitude
			potionPotency = akPotion.GetNthEffectMagnitude(akPotionEffectIndex)
			Debug.TraceConditional(self + " Primary effect potency:" + potionPotency, debugVerbosity)
			While (index + 1) < listLength && potionPotency > (potionsList.GetAt(index) as potion).GetNthEffectMagnitude(closestPotionEffectIndex)
				index = index + 1
				If (potionPotency - closestPotion.GetNthEffectMagnitude(closestPotionEffectIndex)) > math.abs((potionsList.GetAt(index) as potion).GetNthEffectMagnitude(closestPotionEffectIndex) - potionPotency)
					closestPotion = potionsList.GetAt(index) as potion
				EndIf
			EndWhile
		ElseIf closestPotionEffectIndex != -1 && akPotion.GetNthEffectMagicEffect(akPotionEffectIndex).IsEffectFlagSet(0x00400000) ;effect scales with duration
			potionPotency = akPotion.GetNthEffectDuration(akPotionEffectIndex)
			Debug.TraceConditional(self + " Primary effect potency:" + potionPotency, debugVerbosity)
			While (index + 1) < listLength && potionPotency > (potionsList.GetAt(index) as potion).GetNthEffectDuration(closestPotionEffectIndex)
				index = index + 1
				If (potionPotency - closestPotion.GetNthEffectDuration(closestPotionEffectIndex)) > math.abs((potionsList.GetAt(index) as potion).GetNthEffectDuration(closestPotionEffectIndex) - potionPotency)
					closestPotion = potionsList.GetAt(index) as potion
				EndIf
			EndWhile
		EndIf
	EndIf
	
	Debug.TraceConditional(self + " Closest potion with matching effect:" + closestPotion, debugVerbosity)
	return closestPotion
EndFunction

